import contextlib
import errno
import logging
import os
import shutil
import stat
import tempfile

logger = logging.getLogger(__name__)


def handle_remove_readonly(func, path, exc):
    """ Sometimes `shutil.rmtree` needs some help when it cones across read-only
    files or folders. This seems to be especially True when working with git
    repositories on Windows systems. Passing this method as `rmtree`'s `onerror`
    argument will attempt to change permissions on any problem files or folders,
    and retry the failed operation. If the operation still fails the original
    exception will be re-raised.
    """
    exc_value = exc[1]
    if func in (os.rmdir, os.remove, os.unlink):
        if exc_value.errno == errno.EACCES:
            if func is not os.rmdir:
                # make sure parent folder is writable
                parent_path = os.path.dirname(path)
                os.chmod(parent_path, stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO)
            os.chmod(path, stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO)
            func(path)
            return
    raise


@contextlib.contextmanager
def temp_path():
    """ Context manager to create a temporary path and clean it up automatically.

    >>> with temp_path() as tmp:
    ...     os.path.isdir(tmp)
    True
    >>> os.path.isdir(tmp)
    False

    """
    tmp_dir = tempfile.mkdtemp()
    logger.debug(f"created temp folder {tmp_dir}")
    try:
        yield tmp_dir
    finally:
        logger.debug(f"removing temp folder {tmp_dir}")
        shutil.rmtree(tmp_dir, onerror=handle_remove_readonly)


if __name__ == '__main__':
    import doctest
    doctest.testmod()



#!/usr/bin/env python3

import argparse
import os
import sys

import xappt

from xappt.models.parameter import convert


def build_parser() -> argparse.ArgumentParser:
    interface_list = [i[0] for i in xappt.plugin_manager.registered_interfaces()]
    default_interface_name = os.environ.get(xappt.INTERFACE_ENV, xappt.INTERFACE_DEFAULT)

    parser = argparse.ArgumentParser()

    parser.add_argument('-v', '--version', action='store_true',
                        help='Display the version number and build')
    parser.add_argument('-i', '--interface', choices=interface_list, default=default_interface_name,
                        help='Specify the name of the default user interface. '
                             f'This can also be done by setting the environment variable {xappt.INTERFACE_ENV}')
    parser.add_argument('--ui', action='store_true', dest="interactive",
                        help='Run the specified tool interactively using the passed command line arguments, '
                             'otherwise it will be invoked with the default interface.')

    subparsers = parser.add_subparsers(help="Sub command help", dest='command')

    for plugin_name, plugin_class in xappt.plugin_manager.registered_tools():
        plugin_parser = subparsers.add_parser(plugin_class.name().lower(), help=plugin_class.help())
        for parameter in plugin_class.class_parameters():
            setup_args = parameter.param_setup_args
            args, kwargs = convert.to_argument_dict(setup_args)
            plugin_parser.add_argument(*args, **kwargs)

    return parser


def cli_main(*argv) -> int:
    parser = build_parser()
    options = parser.parse_args(args=argv)

    if options.version:
        print(f"xappt {xappt.version_str}")
        return 0

    os.environ.setdefault(xappt.INTERFACE_ENV, options.interface)
    interface_class = xappt.plugin_manager.get_default_interface()
    interface = interface_class()

    if options.command is not None:
        plugin_class = xappt.plugin_manager.get_tool_plugin(options.command)
        plugin_instance = plugin_class(**options.__dict__)
        if options.interactive:
            interface.invoke(plugin_instance)
        else:
            plugin_instance.validate()
            return plugin_instance.execute(interface=interface)
    else:
        parser.print_help()

    return 0


def entry_point() -> int:
    xappt.discover_plugins()
    return cli_main(*sys.argv[1:])


if __name__ == '__main__':
    sys.exit(entry_point())



from xappt.utilities import git_tools
from xappt.utilities.command_runner import CommandRunner, CommandResult
from xappt.utilities.find_python import find_python
from xappt.utilities.path import *
from xappt.utilities.humanize import humanize_bytes, humanize_list
from xappt.utilities.termsize import get_terminal_size



from __future__ import annotations

from typing import Dict, Generator, TYPE_CHECKING

from xappt.models.parameter.meta import ParamMeta
from xappt.models.parameter.model import Parameter, ParameterDescriptor
from xappt.models.plugins.base import BasePlugin
if TYPE_CHECKING:
    from xappt.models.plugins.interface import BaseInterface


class BaseTool(BasePlugin, metaclass=ParamMeta):
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            if key in self._parameters_:
                param = getattr(self, key)
                param.value = param.validate(value)

    @classmethod
    def class_parameters(cls) -> Generator[ParameterDescriptor, None, None]:
        for item in cls._parameters_:
            yield getattr(cls, item)

    def parameters(self) -> Generator[Parameter, None, None]:
        for item in self._parameters_:
            yield getattr(self, item)

    def param_dict(self) -> Dict:
        d = {}
        for p in self.parameters():
            d[p.name] = p.value
        return d

    def execute(self, **kwargs) -> int:
        raise NotImplementedError

    def validate(self):
        for param in self.parameters():
            param.validate(param.value)



APP_NAME = "xappt"

PLUGIN_TYPE_TOOL = 1
PLUGIN_TYPE_INTERFACE = 2

PLUGIN_PREFIX = "xappt-", "xappt_"

PLUGIN_PATH_ENV = "XAPPT_PLUGIN_PATH"
INTERFACE_ENV = "XAPPT_INTERFACE"
INTERFACE_DEFAULT = "stdio"

TRUTHY_STRINGS = ("yes", "y", "true")



from typing import Dict, List, Tuple


def to_argument_dict(parameter_dict: Dict) -> Tuple[List, Dict]:
    positional_args = [f"--{parameter_dict['name']}"]
    short_name = parameter_dict.get('options', {}).get('short_name')
    if short_name is not None:
        positional_args.append(f"-{short_name}")

    keyword_args = {}
    if parameter_dict['data_type'] == bool:
        keyword_args['action'] = 'store_true'
    else:
        keyword_args['action'] = 'store'
        keyword_args['type'] = parameter_dict['data_type']
        keyword_args['choices'] = parameter_dict.get('choices')

    keyword_args['help'] = parameter_dict['description']

    if not parameter_dict['required']:
        keyword_args['default'] = parameter_dict['default']
    else:
        keyword_args['required'] = True

    return positional_args, keyword_args



""" All validation routines should do some basic sanity checking and return
either the original value or a sensibly massaged version of the value.

For example, with ParamInt you can specify a set of string choices. If a string is
passed to the int's validate method it should be checked against the choices,
and the index of that choice should be returned rather than the original string.
"""

import re
from typing import Any

import xappt

from xappt.constants import TRUTHY_STRINGS

from xappt.models.parameter.model import Parameter
from xappt.models.parameter.errors import ParameterValidationError

__all__ = [
    'BaseValidator',
    'ValidateRange',
    'ValidateType',
    'ValidateChoiceInt',
    'ValidateChoiceStr',
    'ValidateDefault',
    'ValidateRequired',
    'ValidateDefaultInt',
    'ValidateBoolFromString',
]

NUMERIC_PATTERN = r"-?\d+(?:\.\d+)?"
BOOL_RE = re.compile(rf"^({'|'.join((NUMERIC_PATTERN, ) + TRUTHY_STRINGS)})$", re.I)


class BaseValidator:
    def __init__(self, param: Parameter):
        self.param = param

    def validate(self, value: Any) -> Any:
        return value


class ValidateRequired(BaseValidator):
    def validate(self, value: Any) -> Any:
        if value is None and self.param.required:
            raise ParameterValidationError(f"Missing required parameter {self.param.name}")
        return value


class ValidateType(BaseValidator):
    def validate(self, value: Any) -> Any:
        try:
            return self.param.data_type(value)
        except BaseException as e:
            raise ParameterValidationError(str(e))


class ValidateDefault(BaseValidator):
    def validate(self, value: Any) -> Any:
        if value is None and self.param.default is not None:
            return self.param.default
        return value


class ValidateDefaultInt(BaseValidator):
    def validate(self, value: Any) -> Any:
        if value is None and self.param.default is not None:
            if self.param.choices is not None:
                return self.param.choices[self.param.default]
            else:
                return self.param.default
        return value


class ValidateRange(BaseValidator):
    def __init__(self,  param: Parameter, minimum: Any, maximum: Any):
        super().__init__(param)
        self.minimum = minimum
        self.maximum = maximum

    def validate(self, value: Any) -> Any:
        if self.minimum is not None and value < self.minimum:
            raise ParameterValidationError(f"Value must be at least {self.minimum}")
        if self.maximum is not None and value > self.maximum:
            raise ParameterValidationError(f"Value must be at most {self.maximum}")
        return value


class ValidateChoiceInt(BaseValidator):
    def validate(self, value: Any) -> int:
        if self.param.choices is not None:
            if value not in self.param.choices:
                raise ParameterValidationError(f"Value must be one of {xappt.humanize_list(self.param.choices)}")
            value = self.param.choices.index(value)
        try:
            return int(value)
        except BaseException as e:
            raise ParameterValidationError(str(e))


class ValidateChoiceStr(BaseValidator):
    def validate(self, value: Any) -> str:
        if self.param.choices is not None:
            if value not in self.param.choices:
                raise ParameterValidationError(f"Value must be one of {xappt.humanize_list(self.param.choices)}")
        return value


class ValidateBoolFromString(BaseValidator):
    def validate(self, value: Any) -> bool:
        if isinstance(value, str):
            return BOOL_RE.match(value) is not None
        return value



import xappt


@xappt.register_plugin
class ExamplePlugin(xappt.BaseTool):
    arg1 = xappt.ParamString(required=True)
    arg2 = xappt.ParamString(required=True)
    arg3 = xappt.ParamString(required=True)

    @classmethod
    def name(cls) -> str:
        return "example"

    @classmethod
    def help(cls) -> str:
        return "A simple command that will just echo the passed in arguments"

    def execute(self, interface: xappt.BaseInterface, **kwargs) -> int:
        print(self.arg1.value)
        print(self.arg2.value)
        print(self.arg3.value)
        return 0



import time

import xappt


@xappt.register_plugin
class Interactive1(xappt.BaseTool):
    param1 = xappt.ParamInt(required=True)

    def execute(self, **kwargs) -> int:
        xappt.interface.progress_start()

        for i in range(1, 101):
            xappt.interface.progress_update(f"Tests are fun: i == {i}", i / 100.0)
            time.sleep(0.01)

        xappt.interface.progress_end()

        xappt.interface.message("progress ended")

        try:
            if xappt.interface.ask("Enter more input?"):
                result = xappt.interface.invoke(Interactive2(), **self.param_dict())
            else:
                xappt.interface.message("Aborting")
                result = 0
        except KeyboardInterrupt:
            result = 1
        return result


@xappt.register_plugin(visible=False)
class Interactive2(xappt.BaseTool):
    param1 = xappt.ParamInt(description="A number between 1 and 10, inclusive", minimum=1, maximum=10)
    param2 = xappt.ParamInt(description="A number between 10 and 20, inclusive", minimum=10, maximum=20)
    param3 = xappt.ParamInt(default=7, description="Any integer")
    param4 = xappt.ParamBool(description="Run silently?")
    param5 = xappt.ParamString(description="Flip a coin", choices=("heads", "tails"))

    def execute(self, **kwargs) -> int:
        xappt.interface.message(f"Got {self.param1.value}, {self.param2.value}, {self.param3.value}, "
                                f"{self.param4.value}, {self.param5.value}")
        return 0



from __future__ import annotations

import copy

from typing import Any, Dict, List, Optional, Sequence, Tuple, Type
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from xappt.models.parameter.validators import BaseValidator


class Parameter:
    def __init__(self, name: str, *, data_type: Type, **kwargs):
        self.name: str = name
        self.data_type: Type = data_type
        self.description: str = kwargs.get('description', "")
        self.default: Any = kwargs['default']
        self.required: bool = kwargs['required']
        self.choices: Optional[Sequence] = kwargs.get('choices')
        self.options: Dict = kwargs.get('options', {})
        self.validators: List[BaseValidator] = []
        self.value = kwargs['value']

        for validator in kwargs.get('validators', []):
            if isinstance(validator, Tuple):
                validator, *args = validator
            else:
                args = []
            self.validators.append(validator(self, *args))

    def validate(self, value: Any) -> Any:
        for validator in self.validators:
            value = validator.validate(value)
        return value


class ParameterDescriptor:
    __counter = 0

    def __init__(self, data_type: Type, **kwargs):
        cls = self.__class__
        self.storage_name = '_{}#{}'.format(cls.__name__, cls.__counter)
        kwargs['data_type'] = data_type
        if 'default' not in kwargs:
            kwargs['required'] = True
        else:
            kwargs['required'] = False
        kwargs['default'] = kwargs.get('default')
        kwargs['value'] = kwargs.get('value', kwargs['default'])
        if "description" not in kwargs:
            kwargs['description'] = ""
        self.param_setup_args = copy.deepcopy(kwargs)
        cls.__counter += 1

    def get_parameter(self, instance):
        param = getattr(instance, self.storage_name, None)
        if param is None:
            # Create new parameter instance of descriptor properties
            param = Parameter(**self.param_setup_args)
            setattr(instance, self.storage_name, param)
            return param
        return param

    def __get__(self, instance, owner):
        if instance is None:
            # Call was not through an instance
            return self
        else:
            return self.get_parameter(instance)

    def __set__(self, instance, value):
        raise AttributeError("Parameter is read only. Use Parameter's value attribute")



from .model import ParameterDescriptor


class ParamMeta(type):
    def __new__(mcs, name, bases, attrs):
        cls = type.__new__(mcs, name, bases, attrs)
        cls._parameters_ = []
        for var_name, var_value in vars(cls).items():
            if isinstance(var_value, ParameterDescriptor):
                cls._parameters_.append(var_name)
                var_value.param_setup_args['name'] = var_name
        return cls
